/* This program interfaces with the arduino nano which in turn interfaces with a fifo-delay-line. The arduino
 * speaks a very simple protocol over a serial port; this program is a simple wrapper for it on the PC.
 * See arduino_delay.ino for the arduino's own source code (running on the ATmega328), which documents the protocol.
 * 
 * Copyright (C) Richard Neill 2012, <arduino at REMOVE.ME.richardneill.org>. This program is Free Software. You can 
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, 
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied. 
 * For the details, please see: http://www.gnu.org/licenses/gpl.html
 */

/* This is the relevant subset of the protocol, copied from "arduino_sketch/arduino_delay/arduino_delay.ino".
 * Each side transfers one line at a time, ending with "\n".  The commands marked with "*" below are the most useful ones.
 * [The Arduino should respond within 1 second in all cases, (measure is slowest), usually much faster.]
 * When compiled with debugging enabled, extra messages begin "DEBUG: ".
 *
 *     PC SENDS         ARDUINO REPLIES                                            ARDUINO DOES
 *     --------         ---------------                                            ------------
 *   * HELLO            DELAY LINE                                                 Nothing: just say hello
 *   * CONFIG           OSC_MHZ: 25, DELAY RANGE: [AAA, BBB] NS, OR 0 TO BYPASS.   Prints configuration.
 *     HELP             COMMANDS: HELLO, CONFIG, HELP, BLINK, BYPASS, ...          This lists the (important) commands: "HELLO; CONFIG; SETDELAY INT_NS; GETDELAY; GETVDELAY; BYPASS; INSERT; MEASURE; COUNT INT_MS; HELP; VERSION; ABOUT; UPTIME; BLINK; RESET"
 *     UPTIME           UP: SSS SECONDS                                            Print uptime in seconds. Wraps after ~ 50 days
 *     BLINK            BLINK                                                      Blink the diagnostic LED.
 *     BYPASS           NOFIFO                                                     Bypasses the FIFO. Data is not delayed at all.
 *     INSERT           FIFO                                                       Inserts the FIFO into the signal path. Data is delayed (this doesn't change the value of the fifo delay).
 *   * GETDELAY         NNN                                                        Current delay in NS. Returns integer number of ns; 0 if bypassed; -1 if uninitialised and not bypassed.
 *     GETVDELAY        WWW, XXX, YYY, ZZZ  #NS, +/-NS, TICKS, MHZ                 Get verbose delay value. Values are respectively: delay(ns), error(ns), ticks, clock(MHz). If FIFO disabled, all values are 0. (-1 for uninitialised).
 *   * SETDELAY DDD     OK                                                         Set new delay value to DDD nanoseconds (coerced by clock ticks). Implicitly inserts the fifo (or bypasses for 0).
 *     SETDELAY EEE     INVALID VALUE: 'EEE'. RANGE IS: [XXX,XXX] NS               Error setting new delay. Value is out of range.
 *     MEASURE          TTT                                                        Measure the delay value (not exact), and print it. Value in ns (resolution ~ 10us). -1 if timed out (at 0.5 second). Doesn't affect the main delay-line.
 *     COUNT    TTT     NNN                                                        Count the number of falling-edges arriving, for TTT ms. The count is exact (for f <~ 2.5 MHz); ms is accurate to about 0.2%.
 *     SLOWSWL          SLOWSWL: 60...   (multiple lines, 1 Hz for 1 minute)       Test mode for the PCB: clocks the FIFO at 40 kHz (1000x slower than normal). Manual entry of data with switch, shown on LED. Ends after 1 minute.
 *     RESET            RESET                                                      Resets the Arduino in software. This also clears the FIFO. (A *Hardware* reset can be done with the serial port DTR line.)
 *     ""               ""                                                         Empty string: does nothing. Returns newline.
 *     [anything]       INVALID COMMAND: 'xxx'. TRY 'HELP'.                        Erroneous command. Returns error message.
 *
 *     VERSION,ABOUT,HELP,TEST,ANALOG0,MAIL,WHY  are also implemented.
 */

/* Todo: lockfile handling needs to be improved. If we find a lock, we can't always just assume it's stale. */

/* Headers */
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <signal.h>
#include <libgen.h>
#include <limits.h>

/* Definitions for how to talk to the Arduino. */
#define SERIAL_PORT	"/dev/arduino_nano_A700fmiq"	/* This arduino-nano should have a dedicated symlink in /dev */
#define LOCKFILE_PREFIX "/var/lock/LCK.."		/* Lockfile begins with this */
#define BAUD_RATE	B115200			/* NB the initial "B", see tcsetattr() */
#define BAUD_RATE_STR   "115200"
#define VTIME_X100MS	1			/* Number of 10ths of a second for serial timeout. (inside polling loop) */
#define MAX_LEN 	128			/* Max length of a string used to communicate. */
#define FIFO_MIN	128			/* Properties of the AL422 FIFO attached to the Arduino */
#define FIFO_MAX	393216
#define XTAL_OSC_MHZ	40			/* The circuit board (not the Arduino) has this crystal for the FIFO */
#define EXT_OSC_MHZ	25			/* External oscillator: connected with the PulseBlaster via LVDS FIN1002, and /4 counter */
#define SOFT_RESET_US   250000			/* Time for Arduino to soft-reset. 50ms is actually enough */
#define HARD_RESET_US   2000000			/* Time for hard-reset. 1.5s is enough; allow 2. */
#define COUNT_MAX_MS	1000000     		/* Max counting time in ms, supported by Arduino (higher would probably cause a serial-communication timeout) */
#define CONNECT_TIMEOUT_MS 5000			/* Timeout for initial connection */
#define CONTINUE_TIMEOUT_MS 2000		/* Timeout for more data to arrive, after starting */
#define SLOWSWL_FACTOR  1000			/* Slowdown in Slowswl mode */

#define USE_ONBOARD_OSC 0        		 /* 1 for onboard 40 MHz osc. module; 0 for external inbound clock, 25 MHz. */
#if (USE_ONBOARD_OSC == 1)
        #define OSC_MHZ    XTAL_OSC_MHZ
#else
        #define OSC_MHZ    EXT_OSC_MHZ
#endif

#define HELLO_CMD	"HELLO"
#define HELLO_RESPONSE  "DELAY LINE"
#define CONFIG_CMD	"CONFIG"
#define UPTIME_CMD	"UPTIME"
#define BLINK_CMD	"BLINK"
#define BLINK_RESPONSE  "BLINK"
#define BYPASS_CMD	"BYPASS"
#define BYPASS_RESPONSE	"NOFIFO"
#define INSERT_CMD	"INSERT"
#define INSERT_RESPONSE "FIFO"
#define GETDELAY_CMD	"GETDELAY"
#define GETVDELAY_CMD   "GETVDELAY"
#define SETDELAY_CMD	"SETDELAY"
#define SETDELAY_RESPONSE "OK"
#define MEASURE_CMD	"MEASURE"
#define COUNT_CMD   	"COUNT"
#define SLOWSWL_CMD	"SLOWSWL"
#define SLOWSWL_RESPONSE "SLOWSWL"
#define RESET_CMD	"RESET"
#define RESET_RESPONSE	"RESET"
#define DEBUG_PREFIX    "DEBUG: "

/* NI4462 constants: for documentation */
#define NI4462_FILTER_DELAY_SAMPLES	63	/* Number of samples in the digital filter group delay */
#define NI4462_REFTRIGGER_SAMPLES	2	/* Min number of reference trigger samples allowed (used as an example). */
#define NI4462_CLOCK_MAX_KHZ		204.8	/* NI4462 max sample freq */

/* Internal constants */
#define ONELINE		0
#define MULTILINE_LAST	1
#define ONELINE_WAIT    2
#define ONELINE_SHORT   3

/* Macros */
#define eprintf(...)	fprintf(stderr, __VA_ARGS__)				/* Error printf: send to stderr  */
#define deprintf(...)	if (debug) { fprintf(stderr, __VA_ARGS__); };		/* Debug error printf: print to stderr iff debug is set.  */
#define feprintf(...)	fprintf(stderr, __VA_ARGS__); unlock_serial_port(); exit (EXIT_FAILURE)	/* Fatal error printf: send to stderr, unlock, and exit */

/* Globals */
int debug = 0;

/* Prototypes where needed */
void lock_serial_port();
void unlock_serial_port();

/* Show help */
void print_help(char *argv0){
	eprintf("This program interfaces with the Arduino Nano to control an AL422 3 Mbit FIFO which acts as a\n"
		"clocked delay-line between the PulseBlaster and the NI4462 PFI0. The FIFO can be either inserted\n"
		"or bypassed, and the length of its delay is controlled by pre-filling it (via the Arduino).\n"
		"The delay length can be 0 (bypassed) or between [%d, %d] ns; it is clocked at a speed\n"
		"of %d MHz, (so there is a jitter of %.2f ns, and a maximum signal frequency of %.1f MHz).\n"
#if (USE_ONBOARD_OSC == 1)
		"The FIFO's clock is a dedicated oscillator module, not synchronised with the NI4462\n"
#else
		"The FIFO's clock is directly exported from the NI4462 (via LVDS and divider), for perfect sync.\n"
#endif
		"The Arduino is connected to serial port %s (specified by a udev rule),\n"
		"which runs at %s Baud, with HUPCL off to prevent accidental DTR-reset on exit.\n"
		"\n"
		"USAGE:    %s  OPTIONS      (all permutations permitted; not all are sensible.)\n"
		"\n"
		"OPTIONS:  -h           show this help.\n"
		"\n"
		"          -d           print debugging messages to stderr (uses timeouts, so slower).\n"
		"\n"
		"          -a           check communication (and protocol) with Arduino and exit .\n"
		"\n"
		"          -g           get the current delay value in ns. 0 means bypassed; -1 uninitialised.\n"
		"\n"
		"          -s  LEN_NS   set the delay in ns. LEN_NS = { 0, [%d, %d] }. 0 means Bypass.\n"
		"                       LEN may have suffixes s,ms,us,ns. Truncation/quantisation: %d ns.\n"
		"                       Setting the delay prefills the FIFO and inserts it into the path.\n"
		"\n"
		"          -v           get verbose status information from the Arduino.\n"
		"\n"
		"          -b           bypass the FIFO (the delay on subsequent re-insertion is unchanged).\n"
		"\n"
		"          -i           inserts the FIFO into the path.\n"
		"\n"
		"          -m           measure the actual FIFO delay (in ns) vs. the Arduino's clock. Returns\n"
		"                       0 when bypassed. Accuracy: 8 us quantisation.\n"
		"                       [This uses a spare FIFO bit, so is non-destructive of fill-level.]\n"
		"\n"
		"          -c  LEN_MS   count falling-edges for (integer) ms. Max freq ~2.5 MHz; accuracy ~0.2%%.\n"
		"                       LEN may also have suffixes of s,ms. Max LEN: %d ms.\n"
		"                       [FIFO is altered during this measurement; state is then restored.]\n"
		"\n"
		"          -u           gets the Arduino uptime in seconds (wraps at 2^32, i.e. 71 days).\n"
		"\n"
		"          -l           blink the diagnostic LED on the Arduino.\n"
		"\n"
		"          -w           60 second test: %d kHz clock, with switch/led to generate/view pulses.\n"
		"                       [FIFO is bypassed during this test; state is then restored.]\n"
		"\n"
		"          -r           soft reset the Arduino (also clears the FIFO).\n"
		"\n"
		"          -R           hard reset the Arduino (hangup DTR; also clears the FIFO).\n"
		"\n"
		"RETURNS:\n"
		"          0            on success.\n"
		"          1            on error.\n"
		"\n"
		"NOTE: * The NI4462 has a digital group delay of %d samples. If pre-trigger samples is %d,\n"
		"        and it is clocked at %.1f kHz, the trigger-delay should ideally be %#.1f ns.\n"
		"      * Serial protocol is: send 1-line command; wait up to %d ms, read 1-line reply. -d drains the\n"
		"        queue every time, reading n-lines till EOF: necessary if Arduino has '#define DEBUG 1'.\n"
		"      * The Arduino's own oscillator is not used for any timing functions (except -m, -c).\n"
		"      * The circuit doesn't use the Arduino's 8 Analog pins; these are available for future expansion.\n"
		"\n"

		,((FIFO_MIN * 1000) / OSC_MHZ), ((FIFO_MAX * 1000) / OSC_MHZ), OSC_MHZ,
		((float)1000 / (OSC_MHZ*2)), ((float)OSC_MHZ/2), SERIAL_PORT, BAUD_RATE_STR, basename(argv0),
		((FIFO_MIN * 1000) / OSC_MHZ), ((FIFO_MAX * 1000) / OSC_MHZ), (1000/OSC_MHZ), COUNT_MAX_MS, ((1000 * OSC_MHZ) / SLOWSWL_FACTOR),
		NI4462_FILTER_DELAY_SAMPLES, NI4462_REFTRIGGER_SAMPLES, NI4462_CLOCK_MAX_KHZ,
		((NI4462_FILTER_DELAY_SAMPLES + NI4462_REFTRIGGER_SAMPLES) * 1.0e6/NI4462_CLOCK_MAX_KHZ), (100 * VTIME_X100MS) );
}

/* Signal handler: handle Ctrl-C. */
/* Todo? implicitly doing a hard_reset of arduino? But the user might not want that, so just give a hint. */
void handle_signal(int signum __attribute__ ((unused)) ){
	signum=0; //silence gcc Wall.
	eprintf (" Cancelled: abandoning communication with Arduino. Note: the Arduino is still running the command; if necessary, hard-reset with '-R'.\n");
	unlock_serial_port();	/* Can't close serial port: we haven't got the variable in scope. just unlock it. */
	exit (EXIT_FAILURE);  /* Port will be implicitly closed on exit. */
}


/* Handle lockfile for the serial port. For simplicity, if we see a lock, we break it. */
void lockunlock_serial_port(int action){	/* 1 to lock, 0 to unlock */
	char *cpath, *bname, *lname;
	char lockbuf[256] = LOCKFILE_PREFIX;
	FILE *lockfile;
	pid_t pid;

	/* Canonicalise the path, get basename, prepend prefix to get the lockfile name. eg "/var/lock/LCK..ttyUSBX" */
	cpath = realpath (SERIAL_PORT, NULL);
	if (cpath == NULL){
		feprintf ("Error: could not canonicalise serial port path %s: %s\n", SERIAL_PORT, strerror(errno));
	}
	bname = basename (cpath);
	lname = strncat(lockbuf, bname, sizeof(lockbuf)-1);

	if (action == 1){ 	/* lock */
		/* Check for lockfile. if present, exit; if not present, create it with our PID as contents */
		deprintf ("Creating lockfile %s for serial port %s.\n", lname, SERIAL_PORT);
		lockfile = fopen (lname, "r");
		if (lockfile){
			eprintf ("Warning: serial port is locked: the lockfile %s already exists.\n", lname);
			/* Here, we ought to read the file, obtain the PID, see if that PID is live, and either exit or remove a stale lock. *
			   BUT, assuming that the user knows what he's doing, assume all locks are stale! */
			eprintf ("Just assuming this lockfile is stale: removing it anyway...\n");
			unlink (lname);
		}

		lockfile = fopen (lname, "w+"); /* write the pid to the lockfile */
		if (lockfile == NULL){
			feprintf ("Error creating lockfile %s: %s\n", lname, strerror(errno));
		}
		pid = getpid();
		fprintf (lockfile, "%d\n", pid);
		fclose (lockfile);

	}else{			/* unlock */
		unlink (lname);
		deprintf ("Deleted lockfile %s for serial port %s.\n", lname, SERIAL_PORT);
	}
}

/* Lock serial port */
void lock_serial_port(){
	lockunlock_serial_port(1);
}

/* Remove lockfile for serial port. */
void unlock_serial_port(){
	static int count = 0;
	count++;		/* Prevent recursion. Serial port is unlocked on fatal error; if the locking process also creates a fatal error... */
	if (count > 1){
		exit(EXIT_FAILURE);
	}
	lockunlock_serial_port(0);
}

/* Open the serial port (by reference) */
void open_serial_port (int *serial_port_fd, FILE **serial_port){
	/* This currently uses Raw mode: reads return as much data as there is, after 500ms of silence.
	   It might be better to use line-buffered "Canonical" mode, but make the read non-blocking */
	struct  termios toptions;	//See man tcgetattr, or http://www.easysw.com/~mike/serial/serial.html#2_3

	lock_serial_port();

	if (-1 == (*serial_port_fd = open (SERIAL_PORT, O_RDWR | O_NOCTTY ))){;  /* File descriptor */
		feprintf ("Fatal Error: couldn't open (RDWR) arduino serial port %s: %s.\n", SERIAL_PORT, strerror(errno));
	}

	if (tcgetattr(*serial_port_fd, &toptions) < 0) {	/* Set up the port */
		feprintf ("Fatal Error: Couldn't get terminal attributes for serial port %s: %s\n", SERIAL_PORT, strerror(errno));
	}

	cfsetispeed(&toptions, BAUD_RATE);	/* Baud */
	cfsetospeed(&toptions, BAUD_RATE);

	toptions.c_cflag &= ~PARENB;		/* 8N1 */
	toptions.c_cflag &= ~CSTOPB;
	toptions.c_cflag &= ~CSIZE;
	toptions.c_cflag |= CS8;

	toptions.c_cflag &= ~CRTSCTS; 		/* no flow control */
	toptions.c_cflag |= CREAD | CLOCAL;  	/* turn on READ & ignore ctrl lines */

	/* IMPORTANT FOR ARDUINO, ELSE WE RESET THE ATMEGA328 BY ACCIDENT! */
	/* Don't hangup modem (DTR Low) on close(). The arduino has DTR wired to its reset line (via capacitor). */
	toptions.c_cflag &= ~HUPCL;

	toptions.c_iflag &= ~(IXON | IXOFF | IXANY); /* turn off s/w flow ctrl */
	toptions.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); /* make raw */
	toptions.c_oflag &= ~OPOST; 			     /* make raw */

	toptions.c_cc[VMIN]  = 0;		/* Timeouts for read(): see: http://unixwiz.net/techtips/termios-vmin-vtime.html */
	toptions.c_cc[VTIME] = VTIME_X100MS; 	/* Read will return if no data arrives for 1 second */

	if( tcsetattr(*serial_port_fd, TCSANOW, &toptions) < 0) { /* Apply the settings */
		feprintf ("Fatal Error: Couldn't set terminal attributes for serial port %s: %s\n", SERIAL_PORT, strerror(errno));
	}
	if (NULL == (*serial_port = fdopen(*serial_port_fd, "w+"))){	/* Get buffered file pointer for the same fd */
		feprintf ("Fatal Error: couldn't fopen (w+) serial port fd %d: %s.\n", *serial_port_fd, strerror(errno));
	}
	deprintf ("Opened serial port: %s at baud rate: %s.\n", SERIAL_PORT, BAUD_RATE_STR);
}

/* Set serial port to hangup on close (HUPCL). This will cause the Arduino to Hardware-reset. */
void hupcl_serial_port (int serial_port_fd){
	struct  termios toptions;
	if( tcgetattr(serial_port_fd, &toptions) < 0) { /* Get the current settings */
		feprintf ("Fatal Error: Couldn't get terminal attributes for serial port %s: %s\n", SERIAL_PORT, strerror(errno));
	}
	toptions.c_cflag |= HUPCL;
	if( tcsetattr(serial_port_fd, TCSANOW, &toptions) < 0) { /* Apply the settings, including the change. */
		feprintf ("Fatal Error: Couldn't set terminal attributes for serial port %s: %s\n", SERIAL_PORT, strerror(errno));
	}
	usleep (10000); /* wait 10ms */
	deprintf ("Reconfigured serial port with DTR-hangup-on-close.\n");
}

/* Close the serial port */
void close_serial_port (int serial_port_fd, FILE *serial_port){
	fclose (serial_port);
	close  (serial_port_fd);
	unlock_serial_port();
	deprintf ("Closed serial port.\n");
}

/* Write a line to the port, read one back. This function adds and trims the "\n". */
/* mode controls how many lines and timeouts. Options: ONELINE, MULTILINE_LAST, ONELINE_WAIT */
void write_read_port (char *write_msg, char *read_buffer, size_t read_size, FILE *serial_port, int mode){
	int 	ret, len, timeouts, lines;
	char 	*retc;
	char    debug_buffer[MAX_LEN];
	int     connect_max_tries = CONNECT_TIMEOUT_MS / (VTIME_X100MS*100);
	int     continue_max_tries = CONTINUE_TIMEOUT_MS / (VTIME_X100MS*100);

	/* Write the string (without trailing \0), and add \n */
	fputs(write_msg, serial_port);
	ret = fputs ("\n", serial_port);
	if (ret == EOF){		/* If this fails with "Illegal Seek", it probably means DEBUG is ON in the Arduino, but OFF here */
		feprintf ("Fatal Error: fputs() failed, with EOF: %s.\n",strerror(errno));
	}
	deprintf ("Sent:        %s\n", ((write_msg[0]=='\0') ? "[blank]" : write_msg) );

	/* Reading can be in 3 different modes:
	 *    ONELINE:    	loop (over 100ms timeout reads) till we get a single line (which doesn't begin with "DEBUG: ").
				Give up after 50 attempts.
	 *    ONELINE_SHORT:    Give up earlier: 1/5 the length of ONELINE, i.e. 10 attempts.
	 *    ONELINE_WAIT:     loop forever, till we get a single line (which doesn't begin with "DEBUG: "). Never give up.
				This might not ever complete.
         *    MULTILINE_LAST:   loop till we get first line (give up after 50 tries). Then, keep reading till 2 seconds of silence.
	                        Return the last line (which doesn't begin "DEBUG: ").
         *    In all cases, lines starting with "DEBUG: " are printed in debug-mode; otherwise ignored. */

	read_buffer[0]='\0'; /* clear buffer */
	timeouts = lines = 0;
	while (1){
		retc = fgets (debug_buffer, sizeof(debug_buffer), serial_port);
		if (retc != NULL){   	/* received a line */
			lines++;
			timeouts = 0;
			if ((len = strlen(debug_buffer)) > 0){ /* trim trailing newline */
				debug_buffer[len-1] = '\0';
			}
			deprintf ("Received:      %s\n", ((debug_buffer[0]=='\0') ? "[blank]" : debug_buffer) );
			if (strncmp(debug_buffer, DEBUG_PREFIX, strlen(DEBUG_PREFIX))){ /* If it doesn't begin "DEBUG" ... */
 				strncpy (read_buffer, debug_buffer, read_size); /* this might be the one we want. */
				if (mode == ONELINE || mode == ONELINE_SHORT || mode == ONELINE_WAIT){
					break;
				}
			}
		}else{			/* timeout */
			timeouts++;
			if (timeouts > connect_max_tries  && mode == ONELINE){
				deprintf ("Reply timeout after %d * %d ms, giving up.\n", connect_max_tries, 100*VTIME_X100MS);
				break;
			}else if (timeouts > (connect_max_tries/5)  && mode == ONELINE_SHORT){
				deprintf ("Reply timeout after %d * %d ms, giving up.\n", (connect_max_tries/5), 20*VTIME_X100MS);
				break;
			}else if (timeouts > continue_max_tries && mode == MULTILINE_LAST && lines > 0){
				deprintf ("Continuation timeout after %d * %d ms, probably finished.\n", continue_max_tries, 100*VTIME_X100MS);
				break;
			}
			/* if ONELINE_WAIT, we will continue polling for ever (or Ctrl-C) */
		}
	}
}


/* Start here */
int main(int argc, char* argv[]){
	int	opt; extern char *optarg; extern int optind, opterr, optopt;       /* getopt */
	char    buffer[MAX_LEN];
	int	serial_port_fd;
	FILE    *serial_port;
	int     bypass=0, check=0, get_length=0, insert=0, blink=0, measure=0, count=0, reset=0, hard_reset=0, get_uptime=0, get_verbose=0, set_length=0, slowswl=0;
	long    length = 0, quantised_length,  count_ms = 0;
	double  dlength = 0, dcount_ms = 0;
	char    *endptr;

	/* Parse options and check for validity */
        if ((argc > 1) && (!strcmp (argv[1], "--help"))) {      /* Support --help, without the full getopt_long */
                print_help(argv[0]);
                exit (EXIT_SUCCESS);
        }
        if (argc == 1){		/* Expect at least ONE option */
		print_help(argv[0]);
                exit (EXIT_SUCCESS);
        }
        while ((opt = getopt(argc, argv, "abdghilmrRuvwc:s:")) != -1) {  /* Getopt */
                switch (opt) {
                        case 'h':                               /* Help */
				print_help(argv[0]);
				exit (EXIT_SUCCESS);
				break;

			case 'd':				/* Debugging enabled */
				debug = 1;
				break;

			case 'a':				/* Check communication and exit */
				check = 1;
				break;

			case 'b':				/* Bypass Fifo */
				bypass = 1;
				break;

			case 'c':				/* Count falling edges for ARG (ms). Allow Suffixes. */
				count = 1;

				dcount_ms = strtod (optarg, &endptr);
				if (optarg == '\0' || dcount_ms < 0){
					feprintf ("Invalid count time '%s'.\n", optarg);
				}else if (!strcmp (endptr, "ms") || (*endptr == '\0') ){
					dcount_ms *= 1;
				}else if (!strcmp (endptr, "s")){
					dcount_ms *= 1e3;
				}else{
					feprintf ("Invalid count time '%s'.\n", optarg);
				}
				if (round(dcount_ms) != dcount_ms){
					feprintf ("Invalid count time '%s': must be an integer number of ms.\n", optarg);
				}else if (round(dcount_ms) > INT_MAX){
					feprintf ("Invalid count time '%s': too large for LONG_MAX.\n", optarg);
				}
				count_ms = round(dcount_ms);

				break;

			case 'g':				/* Get current delay value */
				get_length = 1;
				break;

			case 'i':				/* Insert Fifo */
				insert = 1;
				break;

			case 'l':				/* Blink LED */
				blink = 1;
				break;

			case 'm':				/* Measure time */
				measure = 1;
				break;

			case 'r':				/* Reset the Arduino */
				reset = 1;
				break;

			case 'R':				/* Hard-Reset the Arduino */
				hard_reset = 1;
				break;

			case 'u':				/* Get uptime */
				get_uptime = 1;
				break;

			case 'v':				/* Get verbose info */
				get_verbose = 1;
				break;

			case 'w':				/* Slow switch/led mode */
				slowswl = 1;
				break;

			case 's':				/* Set length (ns). Allow suffixes. */
				set_length = 1;

				dlength = strtod (optarg, &endptr);
				if (optarg == '\0' || dlength < 0){
					feprintf ("Invalid length '%s'.\n", optarg);
				}else if ((!strcmp (endptr, "ns")) || (*endptr == '\0') ){
					dlength *= 1;
				}else if (!strcmp (endptr, "us")){
					dlength *= 1e3;
				}else if (!strcmp (endptr, "ms")){
					dlength *= 1e6;
				}else if (!strcmp (endptr, "s")){
					dlength *= 1e9;
				}else{
					feprintf ("Invalid length '%s'.\n", optarg);
				}
				if (round(dlength) != dlength){
					feprintf ("Invalid length '%s': must be an integer number of ns.\n", optarg);
				}else if (round(dlength) > INT_MAX){
					feprintf ("Invalid length '%s': too large for LONG_MAX.\n", optarg);
				}
				length = round(dlength);

				break;

			default:
				feprintf ("Unrecognised argument %c. Use -h for help.\n", opt);
				break;
		}
	}

	//Set handler for Ctrl-C. */
	signal(SIGINT, handle_signal);

	/* Configure and open the Serial port. Do this by reference, with BOTH a file-descriptor and a file-handle */
	open_serial_port (&serial_port_fd, &serial_port);

	/* Hard-Reset the Arduino using DTR. */
	if (hard_reset){
		eprintf ("Hard-resetting the Arduino (via DTR).\n");
		hupcl_serial_port (serial_port_fd);  /* Set HUPCL */
		close_serial_port (serial_port_fd, serial_port);  /* Close and reopen the port */
		open_serial_port (&serial_port_fd, &serial_port);
		eprintf ("Sleeping %d ms to allow for it to reboot....", (HARD_RESET_US/1000) );
		usleep (HARD_RESET_US);  /* Allow the ATmega328 time to reboot. (includes the bootloader). Takes ~1.5s, allow 2. */
		eprintf ("done.\n");
	}

	/* Send "" and ignore reply (just to clear buffers, if necessary). Timeout here doesn't trigger error; more meaningful on HELLO/DELAY_LINE. */
	write_read_port ("", buffer, sizeof(buffer), serial_port, ONELINE_SHORT);

	/* Check communication (this also identifies that we have the right device and protocol */
	if (check){
		eprintf ("Checking for Arduino (and protocol)...\n");
	}else{
		deprintf ("Checking for Arduino (and protocol)...\n");
	}
	write_read_port (HELLO_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
	if (strcmp(buffer, HELLO_RESPONSE)){
		feprintf ("Fatal error: sent '%s', expected '%s', got '%s'.\n", HELLO_CMD, HELLO_RESPONSE, buffer);
	}
	if (check){  /* If -c, then exit with the right status if we found our Arduino */
		eprintf ("OK.\n");
	        close_serial_port (serial_port_fd, serial_port);
		exit (EXIT_SUCCESS);
	}

	/* Reset the Arduino */
	if (reset){
		eprintf ("Resetting Arduino and Fifo.\n");
		write_read_port (RESET_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		if (strcmp(buffer, RESET_RESPONSE)){
			feprintf ("Fatal error: sent '%s', expected '%s', got '%s'.\n", RESET_CMD, RESET_RESPONSE, buffer);
		}
		usleep (SOFT_RESET_US);  /* Allow the ATmega328 time to reset. 250ms is more than enough (actually 50ms is enough) */
	}

	/* Blink the Debug LED */
	if (blink){
		eprintf ("Blink.\n");
		write_read_port (BLINK_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		if (strcmp(buffer, BLINK_RESPONSE)){
			feprintf ("Fatal error: sent '%s', expected '%s', got '%s'.\n", BLINK_CMD, BLINK_RESPONSE, buffer);
		}
	}

	/* Get uptime. This is useful to verify that we haven't accidentally rebooted via DTR. */
	if (get_uptime){
		eprintf ("Getting uptime...\n");
		write_read_port (UPTIME_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		printf ("%s\n", buffer);  /* print the message, direct from Arduino. */
	}

	/* Set length (in ns) */
	if (set_length){
		quantised_length = length - (length % (1000/OSC_MHZ) );
		eprintf ("Setting delay to %ld ns (will actually be coerced to %ld ns).\n", length, quantised_length);
		snprintf (buffer, sizeof(buffer), "%s %ld", SETDELAY_CMD, length);
		write_read_port (buffer, buffer, sizeof(buffer), serial_port, ONELINE);
		if (strcmp(buffer, SETDELAY_RESPONSE)){   /* Arduino replies "OK" if happy, or with error message if  out of range */
			feprintf ("Fatal error: sent '%s', expected '%s', got '%s'.\n", SETDELAY_CMD, SETDELAY_RESPONSE, buffer);
		}
	}

	/* Bypass the FIFO */
	if (bypass){
		eprintf ("Bypass.\n");
		write_read_port (BYPASS_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		if (strcmp(buffer, BYPASS_RESPONSE)){
			feprintf ("Fatal error: sent '%s', expected '%s', got '%s'.\n", BYPASS_CMD, BYPASS_RESPONSE, buffer);
		}
	}

	/* Insert the FIFO */
	if (insert){
		eprintf ("Insert.\n");
		write_read_port (INSERT_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		if (strcmp(buffer, INSERT_RESPONSE)){
			feprintf ("Fatal error: sent '%s', expected '%s', got '%s'.\n", INSERT_CMD, INSERT_RESPONSE, buffer);
		}
	}

	/* Get current configured length in ns */
	if (get_length){
		eprintf ("Getting length (ns)...\n");
		write_read_port (GETDELAY_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		printf ("%s\n", buffer);  /* print the message, direct from Arduino. */
	}

	/* Get verbose information */
	if (get_verbose){
		eprintf ("Reading config...\n");
		write_read_port (CONFIG_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		printf ("%s\n", buffer);  /* print the message, direct from Arduino. */

		eprintf ("Getting verbose length (delay ns, jitter ns, steps, fifo clock MHz)...\n");
		write_read_port (GETVDELAY_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		printf ("%s\n", buffer);  /* print the message, direct from Arduino. */
	}

	/* Measure the time for the delay. 0 means bypassed; -1 means Arduino measurement timed out. */
	if (measure){  /* Nb if this fails, it can mean VTIME is too low */
		eprintf ("Measuring length (ns)...\n");
		write_read_port (MEASURE_CMD, buffer, sizeof(buffer), serial_port, ONELINE);
		printf ("%s\n", buffer);
	}

	/* Count falling edges (high frequency) for X ms. (during this time, the fifo is inserted; later bypassed if necessary). */
	if (count){
		eprintf ("Counting falling-edges for %ld ms (Ctrl-C to cancel)...\n", count_ms);
		snprintf (buffer, sizeof(buffer), "%s %ld", COUNT_CMD, count_ms);
		write_read_port (buffer, buffer, sizeof(buffer), serial_port, ONELINE_WAIT);  /* Response may take a long time...wait for it. */
		printf ("%s\n", buffer);
	}

	/* Test mode on circuit board: 40 kHz clock and use switch/led. Main signal bypasses fifo during this time (then restored). Non-destructive. */
	if (slowswl){
		eprintf ("Running test mode (switch/led/40kHz) for 60 seconds (Ctrl-C to cancel)...\n");
		write_read_port (SLOWSWL_CMD, buffer, sizeof(buffer), serial_port, MULTILINE_LAST);
		if (strcmp(buffer, SLOWSWL_RESPONSE)){
			feprintf ("Fatal error: sent '%s', expected '%s', got '%s'.\n", SLOWSWL_CMD, SLOWSWL_RESPONSE, buffer);
		}
		eprintf ("End of test.\n");
	}

	/* Close the serial port */
	close_serial_port (serial_port_fd, serial_port);

	return EXIT_SUCCESS;
}
